
use strict;
use warnings;

use Module::Build 0.3601;


my %module_build_args = (
  'build_requires' => {
    'Module::Build' => '0.3601'
  },
  'configure_requires' => {
    'Module::Build' => '0.3601'
  },
  'dist_abstract' => 'Ensure symlink existence and target',
  'dist_author' => [
    'Steven Haryanto <stevenharyanto@gmail.com>'
  ],
  'dist_name' => 'Setup-File-Symlink',
  'dist_version' => '0.09',
  'license' => 'perl',
  'module_name' => 'Setup::File::Symlink',
  'recommends' => {},
  'recursive_test_files' => 1,
  'requires' => {
    'Data::Dump::Partial' => '0',
    'File::Copy::Recursive' => '0',
    'File::Path' => '2.08',
    'File::Slurp' => '0',
    'File::chdir' => '0',
    'Log::Any' => '0',
    'Sub::Spec' => '0.13',
    'Test::More' => '0.96',
    'UUID::Random' => '0',
    'perl' => '5.010000'
  },
  'script_files' => []
);


my $build = Module::Build->new(%module_build_args);

$build->create_build_script;
